pp -M Carp -M Clone -M Config::General -M Cwd -M Data::Dumper -M Digest::MD5 -M File::Basename -M File::Spec::Functions -M File::Temp -M FindBin -M Font::TTF:: -M GD -M GD::Polyline -M Getopt::Long -M IO::File -M List::MoreUtils -M List::Util -M Math::Bezier -M Math::BigFloat -M Math::Round -M Math::VecStat -M Memoize -M POSIX -M Params::Validate -M Pod::Usage -M Readonly -M Regexp::Common -M SVG -M Set::IntSpan -M Statistics::Basic -M Storable -M Sys::Hostname -M Text::Balanced -M Text::Format -M Time::HiRes  -l c:\Dwimperl\c\bin\libgd-2_.dll -l c:\Dwimperl\c\bin\libfreetype-6_.dll -l c:\Dwimperl\c\bin\libpng12-0_.dll -l c:\Dwimperl\c\bin\libpng15-15_.dll -l c:\Dwimperl\c\bin\libXpm_.dll -l c:\Dwimperl\c\bin\libz_.dll -l c:\Dwimperl\c\bin\libiconv-2_.dll -l c:\Dwimperl\c\bin\libjpeg-62_.dll -o circos.exe circos
